@@ -5,7 +5,14 @@
 #include <sys/wait.h>
 #include <glib-unix.h>
 #include <gio/gunixinputstream.h>
+#include <gio/gunixoutputstream.h>
 #include <gio/gfiledescriptorbased.h>
+#include <unistd.h>
+#include <fcntl.h>
+#endif
+
+#ifdef __linux__
+#include <sys/ptrace.h>
 #endif
 
 /* We write 2^1 + 2^2 ... + 2^10 or 2047 copies of "Hello World!\n"
@@ -24,6 +31,14 @@
 #define SPLICELEN (TOTAL_HELLOS * strlen (HELLO_WORLD))
 #endif
 
+
+
+#ifdef G_OS_WIN32
+#define TESTPROG "gsubprocess-testprog.exe"
+#else
+#define TESTPROG "gsubprocess-testprog"
+#endif
+
 static GPtrArray *
 get_test_subprocess_args (const char *mode,
                           ...) G_GNUC_NULL_TERMINATED;
@@ -34,19 +49,12 @@ get_test_subprocess_args (const char *mode,
 {
   GPtrArray *ret;
   char *path;
-  const char *binname;
   va_list args;
   gpointer arg;
 
   ret = g_ptr_array_new_with_free_func (g_free);
 
-#ifdef G_OS_WIN32
-  binname = "gsubprocess-testprog.exe";
-#else
-  binname = "gsubprocess-testprog";
-#endif
-
-  path = g_test_build_filename (G_TEST_BUILT, binname, NULL);
+  path = g_test_build_filename (G_TEST_BUILT, TESTPROG, NULL);
   g_ptr_array_add (ret, path);
   g_ptr_array_add (ret, g_strdup (mode));
 
@@ -165,6 +173,31 @@ test_search_path (void)
 
   g_object_unref (proc);
 }
+
+static void
+test_search_path_from_envp (void)
+{
+  GError *local_error = NULL;
+  GError **error = &local_error;
+  GSubprocessLauncher *launcher;
+  GSubprocess *proc;
+  const char *path;
+
+  path = g_test_get_dir (G_TEST_BUILT);
+
+  launcher = g_subprocess_launcher_new (G_SUBPROCESS_FLAGS_SEARCH_PATH_FROM_ENVP);
+  g_subprocess_launcher_setenv (launcher, "PATH", path, TRUE);
+
+  proc = g_subprocess_launcher_spawn (launcher, error, TESTPROG, "exit1", NULL);
+  g_assert_no_error (local_error);
+  g_object_unref (launcher);
+
+  g_subprocess_wait_check (proc, NULL, error);
+  g_assert_error (local_error, G_SPAWN_EXIT_ERROR, 1);
+  g_clear_error (error);
+
+  g_object_unref (proc);
+}
 #endif
 
 static void
@@ -231,7 +264,7 @@ test_exit1_cancel (void)
   GSubprocess *proc;
   TestExit1CancelData data = { 0 };
 
-  g_test_bug ("786456");
+  g_test_bug ("https://bugzilla.gnome.org/show_bug.cgi?id=786456");
 
   args = get_test_subprocess_args ("exit1", NULL);
   proc = g_subprocess_newv ((const gchar * const *) args->pdata, G_SUBPROCESS_FLAGS_NONE, error);
@@ -286,7 +319,7 @@ test_exit1_cancel_in_cb (void)
   GSubprocess *proc;
   TestExit1CancelData data = { 0 };
 
-  g_test_bug ("786456");
+  g_test_bug ("https://bugzilla.gnome.org/show_bug.cgi?id=786456");
 
   args = get_test_subprocess_args ("exit1", NULL);
   proc = g_subprocess_newv ((const gchar * const *) args->pdata, G_SUBPROCESS_FLAGS_NONE, error);
@@ -609,7 +642,7 @@ on_subprocess_exited (GObject         *object,
           g_propagate_error (&data->error, error);
         }
     }
-  g_spawn_check_exit_status (g_subprocess_get_exit_status (subprocess), &error);
+  g_spawn_check_wait_status (g_subprocess_get_status (subprocess), &error);
   g_assert_no_error (error);
   data->events_pending--;
   if (data->events_pending == 0)
@@ -728,11 +761,15 @@ on_communicate_complete (GObject               *proc,
     {
       if (data->is_utf8)
         {
+          g_assert_nonnull (stdout_str);
           stdout_data = (guint8*)stdout_str;
           stdout_len = strlen (stdout_str);
         }
       else
-        stdout_data = g_bytes_get_data (stdout_bytes, &stdout_len);
+        {
+          g_assert_nonnull (stdout_bytes);
+          stdout_data = g_bytes_get_data (stdout_bytes, &stdout_len);
+        }
 
       g_assert_cmpmem (stdout_data, stdout_len, "# hello world" LINEEND, 13 + strlen (LINEEND));
     }
@@ -769,7 +806,7 @@ test_communicate_async (gconstpointer test_data)
   GSubprocessFlags flags = GPOINTER_TO_INT (test_data);
   GError *error = NULL;
   GPtrArray *args;
-  TestAsyncCommunicateData data = { flags, 0, };
+  TestAsyncCommunicateData data = { flags, 0, 0, NULL };
   GSubprocess *proc;
   GCancellable *cancellable = NULL;
   GBytes *input;
@@ -1020,7 +1057,7 @@ test_communicate_utf8_async (gconstpointer test_data)
   GSubprocessFlags flags = GPOINTER_TO_INT (test_data);
   GError *error = NULL;
   GPtrArray *args;
-  TestAsyncCommunicateData data = { flags, 0, };
+  TestAsyncCommunicateData data = { flags, 0, 0, NULL };
   GSubprocess *proc;
   GCancellable *cancellable = NULL;
 
@@ -1053,7 +1090,7 @@ test_communicate_utf8_cancelled_async (gconstpointer test_data)
   GSubprocessFlags flags = GPOINTER_TO_INT (test_data);
   GError *error = NULL;
   GPtrArray *args;
-  TestAsyncCommunicateData data = { flags, 0, };
+  TestAsyncCommunicateData data = { flags, 0, 0, NULL };
   GSubprocess *proc;
   GCancellable *cancellable = NULL;
 
@@ -1197,7 +1234,7 @@ test_communicate_utf8_async_invalid (void)
   GSubprocessFlags flags = G_SUBPROCESS_FLAGS_STDOUT_PIPE;
   GError *error = NULL;
   GPtrArray *args;
-  TestAsyncCommunicateData data = { flags, 0, };
+  TestAsyncCommunicateData data = { flags, 0, 0, NULL };
   GSubprocess *proc;
   GCancellable *cancellable = NULL;
 
@@ -1255,14 +1292,12 @@ test_communicate_utf8_invalid (void)
   g_object_unref (proc);
 }
 
-static gboolean
+static void
 send_terminate (gpointer   user_data)
 {
   GSubprocess *proc = user_data;
 
   g_subprocess_force_exit (proc);
-
-  return FALSE;
 }
 
 static void
@@ -1279,7 +1314,7 @@ on_request_quit_exited (GObject        *object,
   g_assert_true (g_subprocess_get_if_signaled (subprocess));
   g_assert_cmpint (g_subprocess_get_term_sig (subprocess), ==, 9);
 #endif
-  g_spawn_check_exit_status (g_subprocess_get_status (subprocess), &error);
+  g_spawn_check_wait_status (g_subprocess_get_status (subprocess), &error);
   g_assert_nonnull (error);
   g_clear_error (&error);
 
@@ -1308,7 +1343,7 @@ test_terminate (void)
 
   g_subprocess_wait_async (proc, NULL, on_request_quit_exited, loop);
 
-  g_timeout_add_seconds (3, send_terminate, proc);
+  g_timeout_add_seconds_once (3, send_terminate, proc);
 
   g_main_loop_run (loop);
 
@@ -1317,14 +1352,12 @@ test_terminate (void)
 }
 
 #ifdef G_OS_UNIX
-static gboolean
+static void
 send_signal (gpointer user_data)
 {
   GSubprocess *proc = user_data;
 
   g_subprocess_send_signal (proc, SIGKILL);
-
-  return FALSE;
 }
 
 static void
@@ -1345,7 +1378,7 @@ test_signal (void)
 
   g_subprocess_wait_async (proc, NULL, on_request_quit_exited, loop);
 
-  g_timeout_add_seconds (3, send_signal, proc);
+  g_timeout_add_seconds_once (3, send_signal, proc);
 
   g_main_loop_run (loop);
 
@@ -1448,41 +1481,96 @@ static void
 test_cwd (void)
 {
   GError *local_error = NULL;
-  GError **error = &local_error;
   GSubprocessLauncher *launcher;
   GSubprocess *proc;
   GPtrArray *args;
   GInputStream *stdout_stream;
   gchar *result;
-  const char *basename;
-  gchar *tmp_lineend;
-  const gchar *tmp_lineend_basename;
+  gsize result_len;
+  const gchar *tmpdir = g_get_tmp_dir ();
+  gchar *tmpdir_basename = NULL, *result_basename = NULL;
 
   args = get_test_subprocess_args ("cwd", NULL);
   launcher = g_subprocess_launcher_new (G_SUBPROCESS_FLAGS_STDOUT_PIPE);
   g_subprocess_launcher_set_flags (launcher, G_SUBPROCESS_FLAGS_STDOUT_PIPE);
-  g_subprocess_launcher_set_cwd (launcher, g_get_tmp_dir ());
-  tmp_lineend = g_strdup_printf ("%s%s", g_get_tmp_dir (), LINEEND);
-  tmp_lineend_basename = g_strrstr (tmp_lineend, G_DIR_SEPARATOR_S);
+  g_subprocess_launcher_set_cwd (launcher, tmpdir);
 
-  proc = g_subprocess_launcher_spawnv (launcher, (const char * const *)args->pdata, error);
+  proc = g_subprocess_launcher_spawnv (launcher, (const char * const *)args->pdata, &local_error);
   g_ptr_array_free (args, TRUE);
   g_assert_no_error (local_error);
 
   stdout_stream = g_subprocess_get_stdout_pipe (proc);
 
-  result = splice_to_string (stdout_stream, error);
+  result = splice_to_string (stdout_stream, &local_error);
+  g_assert_no_error (local_error);
+  result_len = strlen (result);
 
-  basename = g_strrstr (result, G_DIR_SEPARATOR_S);
-  g_assert_nonnull (basename);
-  g_assert_cmpstr (basename, ==, tmp_lineend_basename);
-  g_free (tmp_lineend);
+  /* The result should end with a line ending */
+  g_assert_cmpstr (result + result_len - strlen (LINEEND), ==, LINEEND);
+
+  /* Not sure if the testprog guarantees to return an absolute path for the cwd,
+   * so only compare the basenames. */
+  tmpdir_basename = g_path_get_basename (tmpdir);
+  result_basename = g_path_get_basename (g_strstrip (result));
+  g_assert_cmpstr (tmpdir_basename, ==, result_basename);
+  g_free (tmpdir_basename);
+  g_free (result_basename);
 
   g_free (result);
   g_object_unref (proc);
   g_object_unref (launcher);
 }
 #ifdef G_OS_UNIX
+
+static void
+test_subprocess_launcher_close (void)
+{
+  GError *local_error = NULL;
+  GError **error = &local_error;
+  GSubprocessLauncher *launcher;
+  GSubprocess *proc;
+  GPtrArray *args;
+  int fd, fd2;
+  gboolean is_open;
+
+  /* Open two arbitrary FDs. One of them, @fd, will be transferred to the
+   * launcher, and the other’s FD integer will be used as its target FD, giving
+   * the mapping `fd → fd2` if a child process were to be spawned.
+   *
+   * The launcher will then be closed, which should close @fd but *not* @fd2,
+   * as the value of @fd2 is only valid as an FD in a child process. (A child
+   * process is not actually spawned in this test.)
+   */
+  fd = dup (0);
+  fd2 = dup (0);
+  launcher = g_subprocess_launcher_new (G_SUBPROCESS_FLAGS_NONE);
+  g_subprocess_launcher_take_fd (launcher, fd, fd2);
+
+  is_open = fcntl (fd, F_GETFD) != -1;
+  g_assert_true (is_open);
+  is_open = fcntl (fd2, F_GETFD) != -1;
+  g_assert_true (is_open);
+
+  g_subprocess_launcher_close (launcher);
+
+  is_open = fcntl (fd, F_GETFD) != -1;
+  g_assert_false (is_open);
+  is_open = fcntl (fd2, F_GETFD) != -1;
+  g_assert_true (is_open);
+
+  /* Now test that actually trying to spawn the child gives %G_IO_ERROR_CLOSED,
+   * as g_subprocess_launcher_close() has been called. */
+  args = get_test_subprocess_args ("cat", NULL);
+  proc = g_subprocess_launcher_spawnv (launcher, (const gchar * const *) args->pdata, error);
+  g_ptr_array_free (args, TRUE);
+  g_assert_null (proc);
+  g_assert_error (local_error, G_IO_ERROR, G_IO_ERROR_CLOSED);
+  g_clear_error (error);
+
+  close (fd2);
+  g_object_unref (launcher);
+}
+
 static void
 test_stdout_file (void)
 {
@@ -1645,7 +1733,8 @@ test_child_setup (void)
 }
 
 static void
-test_pass_fd (void)
+do_test_pass_fd (GSubprocessFlags     flags,
+                 GSpawnChildSetupFunc child_setup)
 {
   GError *local_error = NULL;
   GError **error = &local_error;
@@ -1661,18 +1750,20 @@ test_pass_fd (void)
   char *basic_fd_str;
   char *needdup_fd_str;
 
-  g_unix_open_pipe (basic_pipefds, FD_CLOEXEC, error);
+  g_unix_open_pipe (basic_pipefds, O_CLOEXEC, error);
   g_assert_no_error (local_error);
-  g_unix_open_pipe (needdup_pipefds, FD_CLOEXEC, error);
+  g_unix_open_pipe (needdup_pipefds, O_CLOEXEC, error);
   g_assert_no_error (local_error);
 
   basic_fd_str = g_strdup_printf ("%d", basic_pipefds[1]);
   needdup_fd_str = g_strdup_printf ("%d", needdup_pipefds[1] + 1);
 
   args = get_test_subprocess_args ("write-to-fds", basic_fd_str, needdup_fd_str, NULL);
-  launcher = g_subprocess_launcher_new (G_SUBPROCESS_FLAGS_NONE);
+  launcher = g_subprocess_launcher_new (flags);
   g_subprocess_launcher_take_fd (launcher, basic_pipefds[1], basic_pipefds[1]);
   g_subprocess_launcher_take_fd (launcher, needdup_pipefds[1], needdup_pipefds[1] + 1);
+  if (child_setup != NULL)
+    g_subprocess_launcher_set_child_setup (launcher, child_setup, NULL, NULL);
   proc = g_subprocess_launcher_spawnv (launcher, (const gchar * const *) args->pdata, error);
   g_ptr_array_free (args, TRUE);
   g_assert_no_error (local_error);
@@ -1702,7 +1793,319 @@ test_pass_fd (void)
   g_object_unref (proc);
 }
 
+static void
+test_pass_fd (void)
+{
+  do_test_pass_fd (G_SUBPROCESS_FLAGS_NONE, NULL);
+}
+
+static void
+empty_child_setup (gpointer user_data)
+{
+}
+
+static void
+test_pass_fd_empty_child_setup (void)
+{
+  /* Using a child setup function forces gspawn to use fork/exec
+   * rather than posix_spawn.
+   */
+  do_test_pass_fd (G_SUBPROCESS_FLAGS_NONE, empty_child_setup);
+}
+
+static void
+test_pass_fd_inherit_fds (void)
+{
+  /* Try to test the optimized posix_spawn codepath instead of
+   * fork/exec. Currently this requires using INHERIT_FDS since gspawn's
+   * posix_spawn codepath does not currently handle closing
+   * non-inherited fds. Note that using INHERIT_FDS means our testing of
+   * g_subprocess_launcher_take_fd() is less-comprehensive than when
+   * using G_SUBPROCESS_FLAGS_NONE.
+   */
+  do_test_pass_fd (G_SUBPROCESS_FLAGS_INHERIT_FDS, NULL);
+}
+
+static void
+do_test_fd_conflation (GSubprocessFlags     flags,
+                       GSpawnChildSetupFunc child_setup,
+                       gboolean             test_child_err_report_fd)
+{
+  char success_message[] = "Yay success!";
+  GError *error = NULL;
+  GOutputStream *output_stream;
+  GSubprocessLauncher *launcher;
+  GSubprocess *proc;
+  GPtrArray *args;
+  int unused_pipefds[2];
+  int pipefds[2];
+  int fd_to_pass_to_child;
+  gsize bytes_written;
+  gboolean success;
+  char *fd_str;
+
+  /* This test must run in a new process because it is extremely sensitive to
+   * order of opened fds.
+   */
+  if (!g_test_subprocess ())
+    {
+      g_test_trap_subprocess (NULL, 0, G_TEST_SUBPROCESS_INHERIT_STDOUT | G_TEST_SUBPROCESS_INHERIT_STDERR);
+      g_test_trap_assert_passed ();
+      return;
+    }
+
+  g_unix_open_pipe (unused_pipefds, O_CLOEXEC, &error);
+  g_assert_no_error (error);
+
+  g_unix_open_pipe (pipefds, O_CLOEXEC, &error);
+  g_assert_no_error (error);
+
+  /* The fds should be sequential since we are in a new process. */
+  g_assert_cmpint (unused_pipefds[0] /* 3 */, ==, unused_pipefds[1] - 1);
+  g_assert_cmpint (unused_pipefds[1] /* 4 */, ==, pipefds[0] - 1);
+  g_assert_cmpint (pipefds[0] /* 5 */, ==, pipefds[1] /* 6 */ - 1);
+
+  /* Because GSubprocess allows arbitrary remapping of fds, it has to be careful
+   * to avoid fd conflation issues, e.g. it should properly handle 5 -> 4 and
+   * 4 -> 5 at the same time. GIO previously attempted to handle this by naively
+   * dup'ing the source fds, but this was not good enough because it was
+   * possible that the dup'ed result could still conflict with one of the target
+   * fds. For example:
+   *
+   * source_fd 5 -> target_fd 9, source_fd 3 -> target_fd 7
+   *
+   * dup(5) -> dup returns 8
+   * dup(3) -> dup returns 9
+   *
+   * After dup'ing, we wind up with: 8 -> 9, 9 -> 7. That means that after we
+   * dup2(8, 9), we have clobbered fd 9 before we dup2(9, 7). The end result is
+   * we have remapped 5 -> 9 as expected, but then remapped 5 -> 7 instead of
+   * 3 -> 7 as the application intended.
+   *
+   * This issue has been fixed in the simplest way possible, by passing a
+   * minimum fd value when using F_DUPFD_CLOEXEC that is higher than any of the
+   * target fds, to guarantee all source fds are different than all target fds,
+   * eliminating any possibility of conflation.
+   *
+   * Anyway, that is why we have the unused_pipefds here. We need to open fds in
+   * a certain order in order to trick older GSubprocess into conflating the
+   * fds. The primary goal of this test is to ensure this particular conflation
+   * issue is not reintroduced. See glib#2503.
+   *
+   * This test also has an alternate mode of operation where it instead tests
+   * for conflation with gspawn's child_err_report_fd, glib#2506.
+   *
+   * Be aware this test is necessarily extremely fragile. To reproduce these
+   * bugs, it relies on internals of gspawn and gmain that will likely change
+   * in the future, eventually causing this test to no longer test the bugs
+   * it was originally designed to test. That is OK! If the test fails, at
+   * least you know *something* is wrong.
+   */
+  if (test_child_err_report_fd)
+    fd_to_pass_to_child = pipefds[1] + 2 /* 8 */;
+  else
+    fd_to_pass_to_child = pipefds[1] + 3 /* 9 */;
+
+  launcher = g_subprocess_launcher_new (flags);
+  g_subprocess_launcher_take_fd (launcher, pipefds[0] /* 5 */, fd_to_pass_to_child);
+  g_subprocess_launcher_take_fd (launcher, unused_pipefds[0] /* 3 */, pipefds[1] + 1 /* 7 */);
+  if (child_setup != NULL)
+    g_subprocess_launcher_set_child_setup (launcher, child_setup, NULL, NULL);
+  fd_str = g_strdup_printf ("%d", fd_to_pass_to_child);
+  args = get_test_subprocess_args ("read-from-fd", fd_str, NULL);
+  proc = g_subprocess_launcher_spawnv (launcher, (const gchar * const *) args->pdata, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (proc);
+  g_ptr_array_free (args, TRUE);
+  g_object_unref (launcher);
+  g_free (fd_str);
+
+  /* Close the read ends of the pipes. */
+  close (unused_pipefds[0]);
+  close (pipefds[0]);
+
+  /* Also close the write end of the unused pipe. */
+  close (unused_pipefds[1]);
+
+  /* If doing our normal test:
+   *
+   * So now pipefds[0] should be inherited into the subprocess as
+   * pipefds[1] + 2, and unused_pipefds[0] should be inherited as
+   * pipefds[1] + 1. We will write to pipefds[1] and the subprocess will verify
+   * that it reads the expected data. But older broken GIO will accidentally
+   * clobber pipefds[1] + 2 with pipefds[1] + 1! This will cause the subprocess
+   * to hang trying to read from the wrong pipe.
+   *
+   * If testing conflation with child_err_report_fd:
+   *
+   * We are actually already done. The real test succeeded if we made it this
+   * far without hanging while spawning the child. But let's continue with our
+   * write and read anyway, to ensure things are good.
+   */
+  output_stream = g_unix_output_stream_new (pipefds[1], TRUE);
+  success = g_output_stream_write_all (output_stream,
+                                       success_message, sizeof (success_message),
+                                       &bytes_written,
+                                       NULL,
+                                       &error);
+  g_assert_no_error (error);
+  g_assert_cmpint (bytes_written, ==, sizeof (success_message));
+  g_assert_true (success);
+  g_object_unref (output_stream);
+
+  success = g_subprocess_wait_check (proc, NULL, &error);
+  g_assert_no_error (error);
+  g_object_unref (proc);
+}
+
+static void
+test_fd_conflation (void)
+{
+  do_test_fd_conflation (G_SUBPROCESS_FLAGS_NONE, NULL, FALSE);
+}
+
+static void
+test_fd_conflation_empty_child_setup (void)
+{
+  /* Using a child setup function forces gspawn to use fork/exec
+   * rather than posix_spawn.
+   */
+  do_test_fd_conflation (G_SUBPROCESS_FLAGS_NONE, empty_child_setup, FALSE);
+}
+
+static void
+test_fd_conflation_inherit_fds (void)
+{
+  /* Try to test the optimized posix_spawn codepath instead of
+   * fork/exec. Currently this requires using INHERIT_FDS since gspawn's
+   * posix_spawn codepath does not currently handle closing
+   * non-inherited fds.
+   */
+  do_test_fd_conflation (G_SUBPROCESS_FLAGS_INHERIT_FDS, NULL, FALSE);
+}
+
+static void
+test_fd_conflation_child_err_report_fd (void)
+{
+  /* Using a child setup function forces gspawn to use fork/exec
+   * rather than posix_spawn.
+   */
+  do_test_fd_conflation (G_SUBPROCESS_FLAGS_NONE, empty_child_setup, TRUE);
+}
+
+#ifdef __linux__
+
+/* Handle ptrace events on @main_child, and assert that when it exits, it does
+ * so with status %EXIT_SUCCESS, rather than signalling. Other than that, this
+ * just calls %PTRACE_CONT for all trace events. */
+static void
+trace_children (pid_t main_child)
+{
+  int wstatus;
+
+  g_assert_no_errno (waitpid (main_child, &wstatus, 0));
+  g_assert_no_errno (ptrace (PTRACE_SETOPTIONS, main_child, NULL,
+                             (PTRACE_O_TRACEFORK |
+#ifdef HAVE_PTRACE_O_EXITKILL
+                              PTRACE_O_EXITKILL |
 #endif
+                              PTRACE_O_TRACEVFORK |
+                              PTRACE_O_TRACECLONE |
+                              PTRACE_O_TRACEEXEC)));
+  g_assert_no_errno (ptrace (PTRACE_CONT, main_child, NULL, 0));
+
+  while (TRUE)
+    {
+      pid_t pid;
+      int wstatus;
+      int stop_signum;
+      int ptrace_event;
+
+      pid = waitpid (-1, &wstatus, 0);
+      if (pid == -1 && errno == ECHILD)
+        break;
+
+      g_assert_cmpint (errno, ==, 0);
+      g_assert_cmpint (pid, >=, 0);
+
+      if (WIFSTOPPED (wstatus))
+        stop_signum = WSTOPSIG (wstatus);
+      else
+        stop_signum = 0;
+
+      switch (stop_signum)
+        {
+        case SIGTRAP:
+          ptrace_event = (wstatus >> 16) & 0xffff;
+          switch (ptrace_event)
+            {
+            case 0:
+              g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, stop_signum));
+              break;
+            default:
+              g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, 0));
+              break;
+            }
+          break;
+        case SIGSTOP:
+          g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, 0));
+          break;
+        default:
+          if (!WIFEXITED (wstatus) && !WIFSIGNALED (wstatus))
+            g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, stop_signum));
+          break;
+        }
+
+      if (pid == main_child)
+        {
+          g_assert_false (WIFSIGNALED (wstatus));
+          if (WIFEXITED (wstatus))
+            {
+              g_assert_cmpint (WEXITSTATUS (wstatus), ==, EXIT_SUCCESS);
+              break;
+            }
+        }
+    }
+}
+
+#endif  /* __linux__ */
+
+static void
+test_exit_status_trapped (void)
+{
+#ifdef __linux__
+  GPtrArray *args = NULL;
+  pid_t test_child;
+#endif
+
+  g_test_summary ("Test that exit status is reported correctly for ptrace()d child processes");
+  g_test_bug ("https://gitlab.gnome.org/GNOME/glib/-/merge_requests/3433");
+
+#ifdef __linux__
+  /* Call fork() directly here, rather than using #GSubprocess, so that we can
+   * safely call waitpid() on it ourselves without interfering with the internals
+   * of #GSubprocess.
+   * See https://gitlab.gnome.org/GNOME/glib/-/merge_requests/3433#note_1749055 */
+  args = get_test_subprocess_args ("sleep-and-kill", NULL);
+  test_child = fork ();
+  if (test_child == 0)
+    {
+      /* Between fork() and exec() we can only call async-signal-safe functions. */
+      if (ptrace (PTRACE_TRACEME, 0, NULL, NULL) < 0)
+        abort ();
+
+      g_assert_no_errno (execvp (args->pdata[0], (char * const *) args->pdata));
+    }
+
+  trace_children (test_child);
+
+  g_clear_pointer (&args, g_ptr_array_unref);
+#else
+  g_test_skip ("ptrace() support for this test is only tested on Linux");
+#endif
+}
+
+#endif  /* G_OS_UNIX */
 
 static void
 test_launcher_environment (void)
@@ -1760,7 +2163,6 @@ main (int argc, char **argv)
   gsize i;
 
   g_test_init (&argc, &argv, NULL);
-  g_test_bug_base ("https://bugzilla.gnome.org/");
 
   g_test_add_func ("/gsubprocess/noop", test_noop);
   g_test_add_func ("/gsubprocess/noop-all-to-null", test_noop_all_to_null);
@@ -1768,6 +2170,7 @@ main (int argc, char **argv)
   g_test_add_func ("/gsubprocess/noop-stdin-inherit", test_noop_stdin_inherit);
 #ifdef G_OS_UNIX
   g_test_add_func ("/gsubprocess/search-path", test_search_path);
+  g_test_add_func ("/gsubprocess/search-path-from-envp", test_search_path_from_envp);
   g_test_add_func ("/gsubprocess/signal", test_signal);
 #endif
   g_test_add_func ("/gsubprocess/exit1", test_exit1);
@@ -1835,10 +2238,18 @@ main (int argc, char **argv)
   g_test_add_func ("/gsubprocess/env/inherit", test_env_inherit);
   g_test_add_func ("/gsubprocess/cwd", test_cwd);
 #ifdef G_OS_UNIX
+  g_test_add_func ("/gsubprocess/launcher-close", test_subprocess_launcher_close);
   g_test_add_func ("/gsubprocess/stdout-file", test_stdout_file);
   g_test_add_func ("/gsubprocess/stdout-fd", test_stdout_fd);
   g_test_add_func ("/gsubprocess/child-setup", test_child_setup);
-  g_test_add_func ("/gsubprocess/pass-fd", test_pass_fd);
+  g_test_add_func ("/gsubprocess/pass-fd/basic", test_pass_fd);
+  g_test_add_func ("/gsubprocess/pass-fd/empty-child-setup", test_pass_fd_empty_child_setup);
+  g_test_add_func ("/gsubprocess/pass-fd/inherit-fds", test_pass_fd_inherit_fds);
+  g_test_add_func ("/gsubprocess/fd-conflation/basic", test_fd_conflation);
+  g_test_add_func ("/gsubprocess/fd-conflation/empty-child-setup", test_fd_conflation_empty_child_setup);
+  g_test_add_func ("/gsubprocess/fd-conflation/inherit-fds", test_fd_conflation_inherit_fds);
+  g_test_add_func ("/gsubprocess/fd-conflation/child-err-report-fd", test_fd_conflation_child_err_report_fd);
+  g_test_add_func ("/gsubprocess/exit-status/trapped", test_exit_status_trapped);
 #endif
   g_test_add_func ("/gsubprocess/launcher-environment", test_launcher_environment);
 
